// $Id: dynatracer.h,v 1.10 2011/07/09 22:30:19 bobgian Exp $

/*
 * Copyright 2009-2010 Bob Giansiracusa, Mary Kuhner, Jon Yamato, and Joseph Felsenstein
 *
 * This software is distributed free of charge for non-commercial use
 * and is copyrighted.  Of course, we do not guarantee that the software
 * works, and are not responsible for any damage you may cause or have.
 */

// The name of this file should be changed to "dynameter.h", but CVS makes that *so* difficult ...
#ifndef DYNATRACER_H
#define DYNATRACER_H

//------------------------------------------------------------------------------------
// Must include this file first, before testing LAMARC_QA_DYNAMETER for definedness,
// because this is the file which defines that symbol!

#include "local_build.h"

//------------------------------------------------------------------------------------

#ifdef LAMARC_QA_DYNAMETER
#include <fstream>                      // Required to define std::ofstream.
#include "dynacount.h"                  // Required to define DYNAMETER_ARRAYSIZE.

//------------------------------------------------------------------------------------

// Used in experimental code - not used in current distribution.
// #define LAMARC_QA_MEMORYTRACE
// #define LAMARC_QA_MEMORYPRINT

//------------------------------------------------------------------------------------
// Definition of experimental class for testing overloading of new and delete.

#ifdef LAMARC_QA_MEMORYTRACE            // Under construction - not running in current distribution.

class S
{
  private:
    int i[100];
  public:
    S() { puts("S::S()"); }
    ~S() { puts("S::~S()"); }
};

#endif // LAMARC_QA_MEMORYTRACE - End of experimental code - not running in current distribution.

//------------------------------------------------------------------------------------
// Experimental test of overloading global and class new.

#ifdef LAMARC_QA_MEMORYTRACE            // Under construction - not running in current distribution.

#define NEW new(__FILE__, __LINE__)
#define DELETE printf("delete: %s at %u\n", __FILE__, __LINE__) ; delete

extern void * operator new (size_t size, char * file, unsigned int line);
extern void * operator new [] (size_t size, char * file, unsigned int line);

#endif // LAMARC_QA_MEMORYTRACE - End of experimental code - not running in current distribution.

//------------------------------------------------------------------------------------
// Experimental test of overloading global and class delete.

#ifdef LAMARC_QA_MEMORYTRACE            // Under construction - not running in current distribution.

extern void operator delete (void * ptr);
extern void operator delete [] (void * ptr);

extern void operator delete (void * ptr, size_t size);
extern void operator delete [] (void * ptr, size_t size);

#endif // LAMARC_QA_MEMORYTRACE - End of experimental code - not running in current distribution.

//------------------------------------------------------------------------------------
// Class which defines accumulation of non-dynamic tracing/metering information (total calls/timing, etc).

class GlobalDynameter
{
    friend class LocalDynameter;

    // Utility sorting functions.
    friend bool NumSelfCalls(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool MaxSinceLastBefore(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool CalleeRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool CalleeRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool SelfRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool SelfRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool TotalRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);
    friend bool TotalRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

  private:
    const char * const m_srcFunctionName;            // Function name decorated with class and argument signature.
    const char * const m_srcFileName;                // Source file name.
    const unsigned int m_ordinalCount;               // This function's order of first call (relative to all others).
    unsigned long long int m_calleeRuntime;          // Accum runtime of functions called by this one (callees).
    unsigned long long int m_selfRuntime;            // Accum runtime of this function (not including callees).
    unsigned long long int m_maxSinceLastBeforeCall; // Max ticks since last metering tool access before current call.
    unsigned long long int m_selfNumberOfCalls;      // Num traced calls to this function (not necessarily printed).

    // Constructor for GlobalDynameter called by LocalDynameter constructor to store per-function results.
    GlobalDynameter(const char * const srcFunctionName,
                    const char * const srcFileName,
                    const unsigned int ordinalCount,
                    const unsigned long long int calleeRuntime,
                    const unsigned long long int selfRuntime,
                    const unsigned long long int maxSinceLastBeforeCall);

    // Default constructor - should never get called.
    GlobalDynameter();

}; // class GlobalDynameter

//------------------------------------------------------------------------------------
// Class which defines dynamic tracing/metering activity (per-call information, dynamic call chain, etc).

class LocalDynameter
{
  private:
    const unsigned long long int m_fcnEntryTime; // Clock count at initial function entry.
    const char * const m_srcFunctionName;        // Function name decorated with class and argument signature.
    const char * const m_srcFileName;            // Source file name.
#if (LAMARC_QA_DYNAMETER >= 3u)
    bool m_tracePrintoutOK;                 // Flag indicating state of trace printout suppression.
    unsigned int m_traceLevel;              // "Local" value (per class object - ie, at given fcn level).
#endif // (LAMARC_QA_DYNAMETER >= 3u)
    unsigned long long int m_calleeRuntime; // Accumulated runtime of functions called by this one (callees).
    GlobalDynameter * m_globalDynameterPtr; // Pointer to GlobalDynameter object in GlobalDynameterArray.
    LocalDynameter * m_parentMeterPtr;      // Pointer to my parent (or NULL).
    unsigned int m_srcLineNumber;           // Source line number where constructor called.

    // "Global" pointer to LocalDynameter object of function currently on top of run-time stack.
    static LocalDynameter * s_currentMeterPtr;

    // Accumulates max time from previous trace recording to current call, over all calls to all functions.
    static unsigned long long int s_maxSinceLastBeforeCall;

    // Total number of function calls traced so far (all functions).
    static unsigned long long int s_totalNumberOfCalls;

    static std::ofstream s_traceOut;    // File stream for tracing/metering output.

    // Array of pointers to GlobalDynameter objects holding trace data.
    static GlobalDynameter * s_GlobalDynameterArray[DYNAMETER_ARRAYSIZE];

    // Number of GlobalDynameter objects populating above array - ie, number of functions traced so far this run.
    // This is used as an ordinal count for GlobalDynameter objects (gives order of each function's first call).
    static unsigned int s_GlobalDynameterCount;

#if (LAMARC_QA_DYNAMETER >= 3u)

    // User-settable limit to depth of tracing based on number of calls to same function.
    static unsigned int s_traceCountLimit;

    // Flag/counter to control whether (and how) ellipsis indication is printed when dynamic tracing is suppressed.
    static unsigned int s_ellipsisCounter;

    // Count of total number of elided tracing printouts.
    static unsigned int s_printoutsElided;

    // Utility function.  Prints marks indicating level of function-call nesting.
    void print_indentation(std::ofstream & stream);

    // Utility function.  Prints message indicating number of elided tracing printouts.
    void print_ellipsis(std::ofstream & stream);

#endif // (LAMARC_QA_DYNAMETER >= 3u)

    // Utility function.  Prints summary statistics on all traced functions.
    void TablePrinter(const std::string heading, const std::string filename);

  public:
    // Set at first call and used as zero-point of elapsed time for run.
    static unsigned long long int s_startClock;

    // Set at first call and updated at end of each call, irrespective of trace level at call.
    static unsigned long long int s_lastClock;

    // Constructor/Printer.
    // Note: If macro __PRETTY_FUNCTION__ (which includes signature and class name in string) is unavailable,
    // use more generally available macro __FUNCTION__ instead (no signature or class name in string).
    LocalDynameter(const char * const srcFunctionName,     // Supplied by __PRETTY_FUNCTION__ macro
                   const char * const srcFileName,         // Supplied by __FILE__ macro
                   const unsigned int srcLineNumber,       // Supplied by __LINE__ macro
                   const unsigned int globalDynameterIdx); // Supplied by DYNACOUNTER_START + __COUNTER__

    // Destructor/Printer.
    ~LocalDynameter();

}; // class LocalDynameter

//------------------------------------------------------------------------------------
// Utility functions for sorting.
// All comparison functions yield sorts in decreasing (non-increasing) order.

// Utility function.  Sorts by number of self calls.
extern bool NumSelfCalls(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by max time since last trace probe before current call.
extern bool MaxSinceLastBefore(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by callee runtime per call (total over all callees).
extern bool CalleeRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by callee runtime aggregate (summed over all calls).
extern bool CalleeRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by self runtime per call (excluding all callees).
extern bool SelfRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by self runtime aggregate (excluding callees, summed over all calls).
extern bool SelfRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by total runtime per call (sum of self plus callee runtime).
extern bool TotalRuntimePerCall(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

// Utility function.  Sorts by total runtime aggregate (sum of self plus callee, summed over all calls).
extern bool TotalRuntimeAggregate(const GlobalDynameter * const x_ptr, const GlobalDynameter * const y_ptr);

//------------------------------------------------------------------------------------

#endif  // LAMARC_QA_DYNAMETER

//------------------------------------------------------------------------------------
// External function declarations.

#if defined(LAMARC_QA_DYNAMETER) || defined(LAMARC_QA_TIMINGTEST)

// Timer utility function (CPU cycle counter).
// This function is used by the Dynameter constructors (in expansion of the StartDynameter() macro)
// and corresponding destructors.  It is also called in a few other places to "reset" the tracing/metering
// clock (ie, after human interaction finishes).
//
// It is also used for timing in versions LAMARC in which the Dynameter subsystem is not present
// (ie, LAMARC_QA_DYNAMETER is not defined); hence the OR in the "#if defined" above to allow its
// definition if LAMARC_QA_TIMINGTEST is defined.
//
extern unsigned long long int rdtsc();

#endif // defined(LAMARC_QA_DYNAMETER) || defined(LAMARC_QA_TIMINGTEST)

//------------------------------------------------------------------------------------
// Macros which invoke tracing/metering functionality.
//
// Note: If predefined macro __PRETTY_FUNCTION__ (which includes signature and class name in string) is unavailable,
// use predefined (and more generally available) macro __FUNCTION__ instead (no signature or class name in string).
//
// Note that they expand into other than ordinary function calls (ie, declarations) if LAMARC_QA_DYNAMETER
// is defined, which is why inline function definitions are insufficient.  If LAMARC_QA_DYNAMETER is NOT
// defined, they expand into whitespace (and therefore are invisible in non-tracing/metering-mode compilations).

#ifdef LAMARC_QA_DYNAMETER

#define StartDynameter()                                         \
    LocalDynameter traceObj(__PRETTY_FUNCTION__,                 \
                            __FILE__,                            \
                            __LINE__,                            \
                            DYNACOUNTER_START + __COUNTER__)

#else // LAMARC_QA_DYNAMETER

#define StartDynameter()

#endif // LAMARC_QA_DYNAMETER

//------------------------------------------------------------------------------------
// Simulation of ASSERT macro, except that it runs whether in debug mode or not.
// These macros are present whether LAMARC_QA_DYNAMETER is defined or not; hence,
// they are available for general debugging (not just when using the metering tool).

#define DebugAssert(condition, message)                                 \
    if(!(condition))                                                    \
    {                                                                   \
        std::cerr << std::endl                                          \
                  << "DebugAssert " << message << ":  " << #condition   \
                  << std::endl                                          \
                  << "Function: " << __PRETTY_FUNCTION__                \
                  << std::endl                                          \
                  << "Filename: " << __FILE__ << " (line " << __LINE__ << ")" \
                  << std::endl << std::endl                             \
                  << "Compiled on " << __DATE__ << " at " << __TIME__   \
                  << std::endl;                                         \
        exit(1);                                                        \
    }

//------------------------------------------------------------------------------------
// Simulation of ASSERT macro, except that it runs whether in debug mode or not.
// Like DebugAssert except that it logs information about where it was called from tracing/metering macro.
// srcLineNumber of ZERO means line number could not be ascertained.
// Caller function name (srcFuncName) can be either Short (__FUNCTION__) or Fancy version (__PRETTY_FUNCTION__).

#define TraceAssert(condition, message, srcFuncName, srcFileName, srcLineNumber) \
    if(!(condition))                                                    \
    {                                                                   \
        std::cerr << std::endl                                          \
                  << "TraceAssert " << message << ":  " << #condition   \
                  << std::endl                                          \
                  << "Main function: " << __PRETTY_FUNCTION__           \
                  << std::endl                                          \
                  << "Main filename: " << __FILE__ << " (line " << __LINE__ << ")" \
                  << std::endl << std::endl                             \
                  << "Caller function: " << srcFuncName                 \
                  << std::endl                                          \
                  << "Caller filename: " << srcFileName << " (line " << srcLineNumber << ")" \
                  << std::endl << std::endl                             \
                  << "Compiled on " << __DATE__ << " at " << __TIME__   \
                  << std::endl;                                         \
        exit(1);                                                        \
    }

//------------------------------------------------------------------------------------

#endif // DYNATRACER_H

//____________________________________________________________________________________
